/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier, SafeRiver/LaBRI

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

extern "C" {
#include "aiger.h"
}

#include "aigprep.h"
#include "zigzag.h"
#include "zigzag_bwd.h"
#include "zigzag_t.h"

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

int         simple_p   = 0;
bool        bmc_only   = 0;
int         repli      = 0;
int         rep_max_sz = INT_MAX;
bool        repGlue    = 0;
bool        bwd        = 0;
bool        trans      = 0;
clock_t     clk_diff, clk;

unsigned UNDEF_LIT = INT_MAX;

SimpSolver* solver;
vec<Lit>    assumptions;

unsigned vars = 0;
std::map<unsigned, ands_def*> map_ands;
std::vector<idef*> mapdef;
std::vector<dll_list<ands_def>*> list_ands;

const char * name;
aiger * model, *model_coi;

int verbose_aigMC, bad_output;

static void SIG_handler(int signum) {
  printf("-1\nb0\n.\n");
  fflush(stdout);
  exit(0);
}

void init_aigMC() {
  model = aiger_init ();
}

void destroy_aigMC() {
  aiger_reset (model);
}

bool isnumeric (const char *str) {
  for (int i = 0; i < strlen(str); i++)
    if (!isdigit(str[i]))
      return 0;
  return 1;
}

const char * usage =
  "usage           : repMC [option] <model> [<start> <stop>]\n"
  "-h              : help\n"
  "-v              : verbose\n"
  "-b              : use output as bad state\n"
  "-simple_path_od : add simple path (loop free) constraints on demand\n"
  "-replicate      : perform simple clause replication\n"
  "-replicate_L    : perform simple clause replication in learnt DB\n"
  "-repG           : replication of glue clause greater than max size\n"
  "-rep_sz <sz>    : max size of replicated clause\n"
  "-bmc            : perform BMC only\n"
  "-bwd            : perform ZigZag backward\n"
  "start           : starting depth (Default: 0) (!! Must always be equal to 0 in this version !!)\n"
  "stop            : max depth      (Default: INT_MAX)\n";

int main (int argc, char ** argv) {
  int i, j, k, start, stop;
  const char * err;
  start = -1;
  stop = -1;
  bool parse_rep_sz = 0;
  if (argc == 1) {
    printf ("%s", usage);
    exit(-1);
  }
  for (i = 1; i < argc; i++) {
    if (parse_rep_sz && isnumeric (argv[i])) {
      rep_max_sz = atoi (argv[i]);
      parse_rep_sz = 0;
    } else if (!strcmp (argv[i], "-h")) {
      printf ("%s", usage);
      exit (0);
    } else if (!strcmp (argv[i], "-v")) {
      verbose_aigMC++;
    } else if (!strcmp (argv[i], "-b")) {
      bad_output = 1;
    } else if (!strcmp (argv[i], "-bmc")) {
      bmc_only = 1;
    } else if (!strcmp (argv[i], "-bwd")) {
      assert(!bwd);
      bwd = 1;
    } else if (!strcmp (argv[i], "-trans")) {
      assert(!trans);
      trans = 1;
    } else if (!strcmp (argv[i], "-simple_path_od")) {
      assert(!simple_p);
      simple_p = 2;
    } else if (!strcmp (argv[i], "-replicate")) {
      assert(!repli);
      repli = 1;
    } else if (!strcmp (argv[i], "-replicate_L")) {
      assert(!repli);
      repli = 2;
    } else if (!strcmp (argv[i], "-repG")) {
      repGlue = 1;
    } else if (!strcmp (argv[i], "-rep_sz")) {
      parse_rep_sz = 1;
    } else if (argv[i][0] == '-') {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (name && stop >= 0) {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (name && !isnumeric (argv[i])) {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (start < 0 && isnumeric (argv[i])) {
      start = atoi (argv[i]);
    } else if (stop < 0 && isnumeric (argv[i])) {
      stop = atoi (argv[i]);
    } else {
      name = argv[i];
    }
  }

  if (start < 0) start = 0;
  if (stop  < 0) stop = INT_MAX;
  
  signal(SIGINT,  SIG_handler);
  signal(SIGTERM, SIG_handler);

  if (verbose_aigMC)
    printf("[MC]  AIGER MC from %d to %d\n", start, stop);
  init_aigMC();

  if (verbose_aigMC)
    printf("[MC]  Checking model : '%s'\n", name);

  
  if (name) err = aiger_open_and_read_from_file (model, name);
  if (err) {
    printf("Parsing error '%s': %s\n", name, err);
    destroy_aigMC();
    return -1;
  }

  if (model->num_justice) {
    printf("Justice properties are not supported yet!\n");
    exit(-1);
  }
  if (model->num_constraints) { 
    printf("Constraints are not supported yet!\n");
    exit(-1);
  }
  if (model->num_fairness) {
    printf("Fairness constraints are not supported yet!\n");
    exit(-1);
  }
  if (!model->num_bad) {
    if (!bad_output) {
      printf("No properties! (add -b to consider output as property)");
      exit(-1);
    } else {
      if (model->num_outputs > 1) {
	printf("Multiple properties are not supported yet!\n");
	exit(-1);

      } else {
	aiger_add_bad (model, model->outputs[0].lit, 0);
      }
    }
  }

  aiger_reencode (model);

  if (verbose_aigMC) printf("[aigprep]  Preprocessing\n");
  destroy_prep();
  init_prep(model, verbose_aigMC);
  iter_defs();
  model = rebuild_aig(true);
  destroy_prep();

  if (verbose_aigMC) {
    printf("[MC]  MILOA = %u %u %u %u %u\n",
	   model->maxvar,
	   model->num_inputs,
	   model->num_latches,
	   model->num_outputs,
	   model->num_ands);
    
    printf("[MC]  BCJF = %u %u %u %u\n",
	   model->num_bad,
	   model->num_constraints,
	   model->num_justice,
	   model->num_fairness);
  }
  
  if(model->bad[0].lit < 2) {
    if (!model->bad[0].lit) {
      if (verbose_aigMC) printf("[MC]  Bad state is equal to FALSE\n");
      printf("0\nb0\n.\n");
    } else {
      if (verbose_aigMC) printf("[MC]  Bad state is equal to TRUE\n");
      printf("1\nb0\n.\n");
    }

  } else {
    try {
      if (bmc_only) {
	if (verbose_aigMC) printf("[MC]  Starting BMC\n");
	assert(!bwd);
	assert(!trans);
	zigzag(start, stop, model, /*elim_hl*/ 1, simple_p, repli, repGlue, rep_max_sz, bmc_only, verbose_aigMC);
      } else {
      if (verbose_aigMC) printf("[MC]  Starting ZigZag\n");
      if (bwd)
	zigzag_bwd(start, stop, model, /*elim_hl*/ 1, simple_p, repli, repGlue, rep_max_sz, verbose_aigMC);
      else if (trans)
	zigzag_t(start, stop, model, /*elim_hl*/ 1, simple_p, repli, repGlue, rep_max_sz, verbose_aigMC);
      else
	zigzag(start, stop, model, /*elim_hl*/ 1, simple_p, repli, repGlue, rep_max_sz, bmc_only, verbose_aigMC);
      }
    } catch(OutOfMemoryException&) {
      printf("-1\nb0\n.\n");
    }
  }


  return 0;
}
